"""
Definition of artifactory base object.
"""
from __future__ import annotations

import requests
from requests import Response

from pyartifactory.models import AuthModel


class ArtifactoryObject:
    """Models the artifactory object."""

    def __init__(self, artifactory: AuthModel) -> None:
        self._artifactory = artifactory
        self._auth = (
            self._artifactory.auth[0],
            self._artifactory.auth[1].get_secret_value(),
        )
        self._api_version = self._artifactory.api_version
        self._verify = self._artifactory.verify
        self._cert = self._artifactory.cert
        self._timeout = self._artifactory.timeout
        self.session = requests.Session()

    def _get(self, route: str, **kwargs) -> Response:
        """
        :param route: API Route
        :param kwargs: Additional parameters to add the request
        :returns  An HTTP response
        """
        return self._generic_http_method_request("get", route, **kwargs)

    def _post(self, route: str, **kwargs) -> Response:
        """
        :param route: API Route
        :param kwargs: Additional parameters to add the request
        :returns  An HTTP response
        """
        return self._generic_http_method_request("post", route, **kwargs)

    def _put(self, route: str, **kwargs) -> Response:
        """
        :param route: API Route
        :param kwargs: Additional parameters to add the request
        :returns  An HTTP response
        """
        return self._generic_http_method_request("put", route, **kwargs)

    def _delete(self, route: str, **kwargs) -> Response:
        """
        :param route: API Route
        :param kwargs: Additional parameters to add the request
        :returns  An HTTP response
        """
        return self._generic_http_method_request("delete", route, **kwargs)

    def _patch(self, route: str, **kwargs) -> Response:
        """
        :param route: API Route
        :param kwargs: Additional parameters to add the request
        :returns  An HTTP response
        """
        return self._generic_http_method_request("patch", route, **kwargs)

    def _generic_http_method_request(
        self,
        method: str,
        route: str,
        raise_for_status: bool = True,
        **kwargs,
    ) -> Response:
        """
        :param method: HTTP method to use
        :param route: API Route
        :param kwargs: Additional parameters to add the request
        :return: An HTTP response
        """

        http_method = getattr(self.session, method)
        response: Response = http_method(
            f"{self._artifactory.url}/{route}",
            auth=self._auth,
            **kwargs,
            verify=self._verify,
            cert=self._cert,
            timeout=self._timeout,
        )
        if raise_for_status:
            response.raise_for_status()
        return response



from __future__ import annotations

import logging
from typing import List, Union, overload

import requests
from requests import Response

from pyartifactory.exception import ArtifactoryError, PermissionAlreadyExistsError, PermissionNotFoundError
from pyartifactory.models import AnyPermission
from pyartifactory.models.auth import AuthModel
from pyartifactory.models.permission import Permission, PermissionV2, SimplePermission
from pyartifactory.objects.object import ArtifactoryObject

logger = logging.getLogger("pyartifactory")


class ArtifactoryPermission(ArtifactoryObject):
    """Models an artifactory permission."""

    def __init__(self, artifactory: AuthModel) -> None:
        super().__init__(artifactory)
        if self._api_version == 2:
            self._uri = "v2/security/permissions"
        if self._api_version == 1:
            self._uri = "security/permissions"

    @overload
    def create(
        self,
        permission: Permission,
    ) -> Permission:
        ...

    @overload
    def create(
        self,
        permission: PermissionV2,
    ) -> PermissionV2:
        ...

    def create(self, permission: AnyPermission) -> AnyPermission:
        """
        Creates a permission
        :param permission: Permission v2 or v1 object
        :return: Permission v2 or v1
        """
        permission_name = permission.name
        try:
            self.get(permission_name)
            logger.debug("Permission %s already exists", permission_name)
            raise PermissionAlreadyExistsError(f"Permission {permission_name} already exists")
        except PermissionNotFoundError:
            self._put(
                f"api/{self._uri}/{permission_name}",
                json=permission.model_dump(by_alias=True),
            )
            logger.debug("Permission %s successfully created", permission_name)
            return self.get(permission_name)

    def get(self, permission_name: str) -> AnyPermission:
        """
        Read permission from artifactory. Fill object if exist
        :param permission_name: Name of the permission to retrieve
        :return: Permission
        """
        try:
            response = self._get(f"api/{self._uri}/{permission_name}")
            logger.debug("Permission %s found", permission_name)
            return (
                Permission(**response.json()) if self._artifactory.api_version == 1 else PermissionV2(**response.json())
            )
        except requests.exceptions.HTTPError as error:
            http_response: Union[Response, None] = error.response
            if isinstance(http_response, Response) and http_response.status_code in (404, 400):
                logger.error("Permission %s does not exist", permission_name)
                raise PermissionNotFoundError(f"Permission {permission_name} does not exist")
            raise ArtifactoryError from error

    def list(self) -> List[SimplePermission]:
        """
        Lists all the permissions
        :return: A list of permissions
        """
        response = self._get(f"api/{self._uri}")
        logger.debug("List all permissions successful")
        return [SimplePermission(**permission) for permission in response.json()]

    @overload
    def update(self, permission: Permission) -> Permission:
        ...

    @overload
    def update(self, permission: PermissionV2) -> PermissionV2:
        ...

    def update(self, permission: AnyPermission) -> AnyPermission:
        """
        Updates an artifactory permission
        :param permission: Permission v2 or v1 object
        :return: Permission v2 or v1
        """
        permission_name = permission.name
        self._put(
            f"api/{self._uri}/{permission_name}",
            json=permission.model_dump(by_alias=True),
        )
        logger.debug("Permission %s successfully updated", permission_name)
        return self.get(permission_name)

    def delete(self, permission_name: str) -> None:
        """
        Removes an artifactory permission
        :param permission_name: Name of the permission to delete
        :return: None
        """
        self.get(permission_name)
        self._delete(f"api/{self._uri}/{permission_name}")
        logger.debug("Permission %s successfully deleted", permission_name)



"""
Definition of all exceptions.
"""
from __future__ import annotations


class ArtifactoryError(Exception):
    """Generic artifactory exception."""


class UserAlreadyExistsError(ArtifactoryError):
    """User already exists."""


class GroupAlreadyExistsError(ArtifactoryError):
    """Group already exists."""


class RepositoryAlreadyExistsError(ArtifactoryError):
    """Repository already exists."""


class PermissionAlreadyExistsError(ArtifactoryError):
    """Permission already exists."""


class UserNotFoundError(ArtifactoryError):
    """The user was not found."""


class GroupNotFoundError(ArtifactoryError):
    """The group was not found."""


class RepositoryNotFoundError(ArtifactoryError):
    """The repository was not found."""


class PermissionNotFoundError(ArtifactoryError):
    """A permission object was not found."""


class ArtifactNotFoundError(ArtifactoryError):
    """An artifact was not found"""


class BadPropertiesError(ArtifactoryError):
    """Property value includes invalid characters"""


class PropertyNotFoundError(ArtifactoryError):
    """All requested properties were not found"""


class InvalidTokenDataError(ArtifactoryError):
    """The token contains invalid data."""


class BuildNotFoundError(ArtifactoryError):
    """Requested build were not found"""



"""
Definition of all build models.
"""
from __future__ import annotations

from typing import Dict, List, Optional

from pydantic import BaseModel, Field


class BuildProperties(BaseModel):
    started: Optional[str] = None
    diff: Optional[str] = None
    project: Optional[str] = None

    # Method to convert model to query string
    def to_query_string(self) -> str:
        # Create a list of key=value pairs for all non-None fields
        properties = [f"{key}={value}" for key, value in self.model_dump(exclude_none=True).items()]

        return "?" + "&".join(properties) if properties else ""


class SimpleBuild(BaseModel):
    """Models an artifactory single build."""

    uri: str
    lastStarted: str


class BuildListResponse(BaseModel):
    """Models all artifactory builds."""

    uri: str = ""
    builds: Optional[List[SimpleBuild]] = None


class Run(BaseModel):
    """Models an artifactory single build run."""

    uri: str
    started: str


class BuildRun(BaseModel):
    """Models artifactory build runs."""

    uri: str
    buildsNumber: Optional[List[Run]] = None


class BuildArtifact(BaseModel):
    """Models artifactory build artifact."""

    type: str
    sha1: str
    sha256: str
    md5: str
    name: str
    path: str


class BuildModules(BaseModel):
    """Models artifactory's build modules."""

    properties: Dict[str, str]
    type: str
    id: str
    artifacts: List[BuildArtifact]


class BuildAgent(BaseModel):
    name: str = ""
    version: str = ""


class Vcs(BaseModel):
    revision: Optional[str] = None
    message: Optional[str] = None
    branch: Optional[str] = None
    url: Optional[str] = None


class BuildInfoDetail(BaseModel):
    """Models artifactory buildInfo dict."""

    properties: Optional[Dict[str, str]] = None
    version: str = ""
    name: str = ""
    number: str = ""
    buildAgent: BuildAgent = BuildAgent()
    agent: BuildAgent = BuildAgent()
    started: str = ""
    durationMillis: int = 0
    artifactoryPrincipal: str = ""
    vcs: Optional[List[Vcs]] = None
    modules: List[BuildModules] = []


class BuildInfo(BaseModel):
    """Models artifactory build."""

    uri: str = ""
    buildInfo: BuildInfoDetail = BuildInfoDetail()


class BuildPromotionResult(BaseModel):
    messages: List[Dict[str, str]] = []


class BuildPromotionRequest(BaseModel):
    status: str = ""
    comment: str = ""
    ciUser: str = ""
    timestamp: str = ""
    dryRun: bool = False
    sourceRepo: str
    targetRepo: str
    copyArtifact: bool = Field(False, alias="copy")
    artifacts: bool = True
    dependencies: bool = False
    scopes: List[str] = []
    properties: Dict[str, List[str]] = {}
    failFast: bool = True


class BuildDeleteRequest(BaseModel):
    project: str = ""
    buildName: str
    buildNumbers: List[str]
    deleteArtifacts: bool = False
    deleteAll: bool = False


class BuildDiffResponseDetail(BaseModel):
    updated: List[str] = []
    unchanged: List[str] = []
    removed: List[str] = []
    new: List[str] = []


class BuildDiffResponse(BaseModel):
    artifacts: BuildDiffResponseDetail = BuildDiffResponseDetail()
    dependencies: BuildDiffResponseDetail = BuildDiffResponseDetail()
    properties: BuildDiffResponseDetail = BuildDiffResponseDetail()


class BuildErrorDetail(BaseModel):
    status: int = 0
    message: str = ""


class BuildError(BaseModel):
    errors: List[BuildErrorDetail]

    # Method to extract error message
    def to_error_message(self) -> str:
        return "\n".join([f"Error status: {_error.status} - {_error.message}" for _error in self.errors])


class LicenseControl(BaseModel):
    """Build's license."""

    runChecks: Optional[bool] = None
    includePublishedArtifacts: Optional[bool] = None
    autoDiscover: Optional[bool] = None
    scopesList: Optional[str] = None
    licenseViolationsRecipientsList: Optional[str] = None


class BuildRetention(BaseModel):
    """Build's retention."""

    deleteBuildArtifacts: Optional[bool] = None
    count: Optional[int] = None
    minimumBuildDate: Optional[int] = None
    buildNumbersNotToBeDiscarded: Optional[List[str]] = None


class Artifacts(BaseModel):
    """Build's artifacts."""

    type: Optional[str] = None
    sha1: Optional[str] = None
    md5: Optional[str] = None
    name: Optional[str] = None


class Dependencies(BaseModel):
    """Build's dependency."""

    type: Optional[str] = None
    sha1: Optional[str] = None
    md5: Optional[str] = None
    id: Optional[str] = None
    scopes: Optional[List[str]] = None


class BuildSingleModule(BaseModel):
    """Build's module model."""

    properties: Optional[Dict[str, str]] = None
    id: Optional[str] = None
    artifacts: Optional[List[Artifacts]] = None
    dependencies: Optional[List[Dependencies]] = None


class Tracker(BaseModel):
    """Build's issue tracker."""

    name: Optional[str] = None
    version: Optional[str] = None


class AffectedIssue(BaseModel):
    """Build's affected issue."""

    key: Optional[str] = None
    url: Optional[str] = None
    summary: Optional[str] = None
    aggregated: Optional[bool] = None


class Issue(BaseModel):
    """Build's issue."""

    tracker: Optional[Tracker] = None
    aggregateBuildIssues: Optional[bool] = None
    aggregationBuildStatus: Optional[str] = None
    affectedIssues: Optional[List[AffectedIssue]] = None


class BuildCreateRequest(BaseModel):
    """Models artifactory build creation data."""

    properties: Dict[str, str] = {}
    version: str = "1.0.1"
    name: str
    number: str
    type: Optional[str] = None
    buildAgent: Optional[BuildAgent] = None
    agent: Optional[BuildAgent] = None
    started: str
    artifactoryPluginVersion: Optional[str] = None
    durationMillis: Optional[int] = None
    artifactoryPrincipal: Optional[str] = None
    url: Optional[str] = None
    vcs: Optional[List[Vcs]] = None
    licenseControl: Optional[LicenseControl] = None
    buildRetention: Optional[BuildRetention] = None
    modules: Optional[List[BuildSingleModule]] = None
    issues: Optional[Issue] = None



"""
Definition of all artifact models.
"""
from __future__ import annotations

import hashlib
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, List, Literal, Optional, Union

from pydantic import BaseModel


class Checksums(BaseModel):
    """Models a checksum."""

    sha1: str
    md5: str
    sha256: str

    @classmethod
    def generate(cls, file_: Path) -> Checksums:
        block_size: int = 65536
        mapping: dict[str, Callable[[], Any]] = {"md5": hashlib.md5, "sha1": hashlib.sha1, "sha256": hashlib.sha256}
        results = {}

        for algorithm, hashing_function in mapping.items():
            hasher = hashing_function()
            with file_.absolute().open("rb") as fd:
                buf = fd.read(block_size)
                while len(buf) > 0:
                    hasher.update(buf)
                    buf = fd.read(block_size)
            results[algorithm] = hasher.hexdigest()

        return cls(**results)


class OriginalChecksums(BaseModel):
    """Models original checksums."""

    sha256: str


class Child(BaseModel):
    """Models a child folder."""

    uri: str
    folder: bool


class ArtifactPropertiesResponse(BaseModel):
    """Models an artifact properties response."""

    uri: str
    properties: Dict[str, List[str]]


class ArtifactInfoResponseBase(BaseModel):
    """The base information available for both file and folder"""

    repo: str
    path: str
    created: Optional[datetime] = None
    createdBy: Optional[str] = None
    lastModified: Optional[datetime] = None
    modifiedBy: Optional[str] = None
    lastUpdated: Optional[datetime] = None
    uri: str


class ArtifactFolderInfoResponse(ArtifactInfoResponseBase):
    """Models an artifact folder info response."""

    children: List[Child]


class ArtifactFileInfoResponse(ArtifactInfoResponseBase):
    """Models an artifact file info response."""

    downloadUri: Optional[str] = None
    remoteUrl: Optional[str] = None
    mimeType: Optional[str] = None
    size: Optional[int] = None
    checksums: Optional[Checksums] = None
    originalChecksums: Optional[OriginalChecksums] = None


class ArtifactListEntryResponse(BaseModel):
    """Base model for an entry in an artifact list response."""

    uri: str
    size: int
    lastModified: datetime
    folder: bool


class ArtifactListFolderResponse(ArtifactListEntryResponse):
    """Models a folder in an artifact list response."""

    folder: Literal[True]


class ArtifactListFileResponse(ArtifactListEntryResponse):
    """Models a file in an artifact list response."""

    folder: Literal[False]
    sha1: Optional[str] = None
    sha2: Optional[str] = None


class ArtifactListResponse(BaseModel):
    """Models an artifact list response."""

    uri: str
    created: datetime
    files: List[Union[ArtifactListFileResponse, ArtifactListFolderResponse]]


class ArtifactStatsResponse(BaseModel):
    """Models an artifact statistics response."""

    uri: str
    downloadCount: int
    lastDownloaded: int
    lastDownloadedBy: Optional[str] = None
    remoteDownloadCount: int
    remoteLastDownloaded: int


ArtifactInfoResponse = Union[ArtifactFolderInfoResponse, ArtifactFileInfoResponse]



from __future__ import annotations

import logging
from typing import Dict, Optional

from pyartifactory.exception import InvalidTokenDataError
from pyartifactory.models.auth import AccessTokenModel, ApiKeyModel, PasswordModel
from pyartifactory.objects.object import ArtifactoryObject

logger = logging.getLogger("pyartifactory")


class ArtifactorySecurity(ArtifactoryObject):
    """Models artifactory security."""

    _uri = "security"

    def get_encrypted_password(self) -> PasswordModel:
        """
        Get the encrypted password of the authenticated requestor.
        :return: str
        """
        response = self._get(f"api/{self._uri}/encryptedPassword")
        logger.debug("Encrypted password successfully delivered")
        return PasswordModel(**response.json())

    def create_access_token(
        self,
        user_name: str,
        expires_in: int = 3600,
        refreshable: bool = False,
        groups: Optional[str] = None,
    ) -> AccessTokenModel:
        """
        Creates an access token.

        :param user_name: Name of the user to whom an access key should be granted. transient token
                          is created if user doesn't exist in artifactory.
        :param expires_in: Expiry time for the token in seconds. For eternal tokens specify 0.
        :param refreshable: If set to true token can be refreshed using the refresh token returned.
        :param groups: A list of groups the token has membership of.
                       If an existing user in artifactory is used with existing memberships
                       groups are automatically implied without specification.
        :return: AccessToken
        """
        payload = {
            "username": user_name,
            "expires_in": expires_in,
            "refreshable": refreshable,
        }
        if groups:
            if not isinstance(groups, list):
                raise ValueError(groups)
            scope = f'member-of-groups:"{",".join(groups)}"'
            payload.update({"scope": scope})
        response = self._post(f"api/{self._uri}/token", data=payload, raise_for_status=False)
        if response.ok:
            return AccessTokenModel(**response.json())
        raise InvalidTokenDataError(response.json().get("error_description", "Unknown error"))

    def revoke_access_token(self, token: Optional[str] = None, token_id: Optional[str] = None) -> bool:
        """
        Revokes an access token.

        :param token: The token to revoke
        :param token_id: The id of a token to revoke
        :return: bool True or False indicating success or failure of token revocation attempt.
        """
        if not any([token, token_id]):
            logger.error("Neither a token or a token id was specified")
            raise InvalidTokenDataError
        payload: Dict[str, Optional[str]] = {"token": token} if token else {"token_id": token_id}
        response = self._post(f"api/{self._uri}/token/revoke", data=payload, raise_for_status=False)
        if response.ok:
            logger.debug("Token revoked successfully, or token did not exist")
            return True
        logger.error("Token revocation unsuccessful, response was %s", response.text)
        return False

    def create_api_key(self) -> ApiKeyModel:
        """
        Create an API key for the current user.
        :return: Error if API key already exists - use regenerate API key instead.
        """
        response = self._post(f"api/{self._uri}/apiKey")
        logger.debug("API Key successfully created")
        return ApiKeyModel(**response.json())

    def regenerate_api_key(self) -> ApiKeyModel:
        """
        Regenerate an API key for the current user
        :return: API key
        """
        response = self._put(f"api/{self._uri}/apiKey")
        logger.debug("API Key successfully regenerated")
        return ApiKeyModel(**response.json())

    def get_api_key(self) -> ApiKeyModel:
        """
        Get the current user's own API key
        :return: API key
        """
        response = self._get(f"api/{self._uri}/apiKey")
        logger.debug("API Key successfully delivered")
        return ApiKeyModel(**response.json())

    def revoke_api_key(self) -> None:
        """
        Revokes the current user's API key
        :return: None
        """
        self._delete(f"api/{self._uri}/apiKey")
        logger.debug("API Key successfully revoked")

    def revoke_user_api_key(self, name: str) -> None:
        """
        Revokes the API key of another user
        :param name: name of the user to whom api key has to be revoked
        :return: None
        """
        self._delete(f"api/{self._uri}/apiKey/{name}")
        logger.debug("User API Key successfully revoked")



"""
Definition of all user related models.
"""
from __future__ import annotations

from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel, EmailStr, SecretStr


class SimpleUser(BaseModel):
    """Models a simple user."""

    name: str
    uri: str
    realm: Optional[str] = None


class BaseUserModel(BaseModel):
    """
    Models a base user.
    https://www.jfrog.com/confluence/display/JFROG/Security+Configuration+JSON#SecurityConfigurationJSON-application/vnd.org.jfrog.artifactory.security.User+json
    """

    name: str
    admin: Optional[bool] = False
    profileUpdatable: Optional[bool] = True
    disableUIAccess: Optional[bool] = False
    internalPasswordDisabled: Optional[bool] = False
    groups: Optional[List[str]] = None


class NewUser(BaseUserModel):
    """Models a new user."""

    email: EmailStr
    password: SecretStr


class User(BaseUserModel):
    """Models a user."""

    email: Optional[EmailStr] = None


class UserResponse(BaseUserModel):
    """Models a user response."""

    email: EmailStr
    lastLoggedIn: Optional[datetime] = None
    realm: Optional[str] = None
    offlineMode: bool = False



"""
Import all object definitions here.
"""
from __future__ import annotations

import contextlib
from importlib.metadata import PackageNotFoundError, version

from pyartifactory.models.auth import AccessTokenModel
from pyartifactory.objects.artifact import ArtifactoryArtifact
from pyartifactory.objects.artifactory import Artifactory
from pyartifactory.objects.build import ArtifactoryBuild
from pyartifactory.objects.group import ArtifactoryGroup
from pyartifactory.objects.permission import ArtifactoryPermission
from pyartifactory.objects.repository import ArtifactoryRepository
from pyartifactory.objects.security import ArtifactorySecurity
from pyartifactory.objects.user import ArtifactoryUser

__all__ = [
    "AccessTokenModel",
    "Artifactory",
    "ArtifactoryGroup",
    "ArtifactoryArtifact",
    "ArtifactoryPermission",
    "ArtifactoryRepository",
    "ArtifactorySecurity",
    "ArtifactoryUser",
    "ArtifactoryBuild",
]

with contextlib.suppress(PackageNotFoundError):
    __version__ = version("pyartifactory")



from __future__ import annotations

import logging
from typing import Union

import requests
from requests import Response

from pyartifactory.exception import ArtifactoryError, BuildNotFoundError
from pyartifactory.models.build import (
    BuildCreateRequest,
    BuildDeleteRequest,
    BuildDiffResponse,
    BuildError,
    BuildInfo,
    BuildListResponse,
    BuildPromotionRequest,
    BuildPromotionResult,
    BuildProperties,
)
from pyartifactory.objects.object import ArtifactoryObject

logger = logging.getLogger("pyartifactory")


class ArtifactoryBuild(ArtifactoryObject):
    """Models an artifactory build."""

    _uri = "build"

    def get_build_info(
        self,
        build_name: str,
        build_number: str,
        properties: BuildProperties = BuildProperties(),
    ) -> BuildInfo:
        """
        :param build_name: Build name to be retrieved
        :param build_number: Build number to be retrieved
        :param properties: Build properties model, for admitted values
                           see https://jfrog.com/help/r/jfrog-rest-apis/build-info
        :return: BuildInfo model object containing server response
        """
        try:
            response = self._get(
                f"api/{self._uri}/{build_name}/{build_number}{properties.to_query_string()}",
            )
            logger.debug("Build Info successfully retrieved")
        except requests.exceptions.HTTPError as error:
            self._raise_exception(error)

        return BuildInfo(**response.json())

    def create_build(self, create_build_request: BuildCreateRequest) -> None:
        try:
            self.get_build_info(create_build_request.name, create_build_request.number)
        except BuildNotFoundError:
            # other exception from get_build_info are forwarded to caller.
            try:
                # build does not exist, can be created here
                self._put(f"api/{self._uri}", json=create_build_request.model_dump())
                logging.debug(
                    "Build %s in %s successfully created",
                    create_build_request.number,
                    create_build_request.name,
                )
            except requests.exceptions.HTTPError as error:
                self._raise_exception(error)
        else:
            logger.error("Build %s in %s already exists", create_build_request.number, create_build_request.name)
            raise ArtifactoryError(f"Build {create_build_request.number} in {create_build_request.name} already exists")

    def promote_build(
        self,
        build_name: str,
        build_number: str,
        promotion_request: BuildPromotionRequest,
    ) -> BuildPromotionResult:
        """
        :param build_name: Build name to be promoted
        :param build_number: Build number to be promoted
        :param promotion_request: Model object containing parameters for promotion
        :return: BuildPromotionResponse containing server response
        """
        try:
            self._get(
                f"api/{self._uri}/{build_name}/{build_number}",
            )
        except requests.exceptions.HTTPError as error:
            self._raise_exception(error)
        else:
            try:
                response = self._post(
                    f"api/{self._uri}/promote/{build_name}/{build_number}",
                    json=promotion_request.model_dump(),
                )
                logging.debug(
                    "Build %s in %s promoted from %s to %s",
                    build_number,
                    build_name,
                    promotion_request.sourceRepo,
                    promotion_request.targetRepo,
                )
            except requests.exceptions.HTTPError as error:
                self._raise_exception(error)

        return BuildPromotionResult(**response.json())

    def list(self) -> BuildListResponse:
        """
        :return: BuildListResponse model object containing server response
        """
        response = self._get(f"api/{self._uri}")
        logger.debug("List all builds successful")
        return BuildListResponse.model_validate(response.json())

    def delete(self, delete_build: BuildDeleteRequest) -> None:
        """
        :param delete_build: Model object containing required parameters
        :return: None
        """
        try:
            for _build_number in delete_build.buildNumbers:
                self._get(
                    f"api/{self._uri}/{delete_build.buildName}/{_build_number}",
                )
            # all build numbers exist
            self._post(f"api/{self._uri}/delete", json=delete_build.model_dump())
            logger.debug("Builds %s deleted from %s", ",".join(delete_build.buildNumbers), delete_build.buildName)
        except requests.exceptions.HTTPError as error:
            self._raise_exception(error)

    def build_rename(self, build_name: str, new_build_name: str) -> None:
        """
        :param build_name: Build to be renamed
        :param new_build_name: New Build name
        :return: None
        """
        try:
            self._get(
                f"api/{self._uri}/{build_name}",
            )
        except requests.exceptions.HTTPError as error:
            self._raise_exception(error)
        else:
            try:
                self._post(f"api/{self._uri}/rename/{build_name}?to={new_build_name}")
                logger.debug("Build %s successfully renamed to %s", build_name, new_build_name)
            except requests.exceptions.HTTPError as error:
                self._raise_exception(error)

    def build_diff(self, build_name: str, build_number: str, older_build_number: str) -> BuildDiffResponse:
        """
        :param build_name: Build name to be compared
        :param build_number: More recent build to be compared
        :param older_build_number: Starting build to be compared
        :return: BuildDiffResponse model object containing server response
        """
        try:
            response = self._get(
                f"api/{self._uri}/{build_name}/{build_number}?diff={older_build_number}",
            )
            logger.debug("Build Diff successfully retrieved between %s and %s", build_number, older_build_number)
        except requests.exceptions.HTTPError as error:
            self._raise_exception(error)

        return BuildDiffResponse(**response.json())

    def _raise_exception(self, error: requests.exceptions.HTTPError):
        http_response: Union[Response, None] = error.response
        if isinstance(http_response, Response):
            _http_error = BuildError(**http_response.json())
            if http_response.status_code == 404:
                raise BuildNotFoundError(_http_error.to_error_message()) from error
            raise ArtifactoryError(_http_error.to_error_message()) from error
        else:
            raise ArtifactoryError from error



from __future__ import annotations

import json
import logging
from typing import List, Union, overload

import requests
from requests import Response

from pyartifactory.exception import ArtifactoryError, RepositoryAlreadyExistsError, RepositoryNotFoundError
from pyartifactory.models import AnyRepository, AnyRepositoryResponse
from pyartifactory.models.repository import (
    FederatedRepository,
    FederatedRepositoryResponse,
    LocalRepository,
    LocalRepositoryResponse,
    RClassEnum,
    RemoteRepository,
    RemoteRepositoryResponse,
    SimpleRepository,
    VirtualRepository,
    VirtualRepositoryResponse,
)
from pyartifactory.objects.object import ArtifactoryObject
from pyartifactory.utils import custom_encoder

logger = logging.getLogger("pyartifactory")


class ArtifactoryRepository(ArtifactoryObject):
    """Models an artifactory repository."""

    _uri = "repositories"

    # Repositories operations
    def get_repo(self, repo_name: str) -> AnyRepositoryResponse:
        """
        Finds repository in artifactory. Raises an exception if the repo doesn't exist.
        :param repo_name: Name of the repository to retrieve
        :return: Either a local, virtual, remote or federated repository
        """
        try:
            response = self._get(f"api/{self._uri}/{repo_name}")
            response_data = response.json()
            rclass = None

            try:
                rclass = response_data["rclass"]
            except KeyError:
                raise KeyError('"rclass" key not found in the response data received by artifactory.')

            # Match to the correct repository type depending on the rclass
            if rclass == RClassEnum.local:
                return LocalRepositoryResponse.model_validate(response_data)
            elif rclass == RClassEnum.virtual:
                return VirtualRepositoryResponse.model_validate(response_data)
            elif rclass == RClassEnum.remote:
                return RemoteRepositoryResponse.model_validate(response_data)
            elif rclass == RClassEnum.federated:
                return FederatedRepositoryResponse.model_validate(response_data)
            else:
                # this should never happen and is a missing repotype in the library
                raise ArtifactoryError(
                    f"Unknown repository type found in response: {rclass}. Please report this issue.",
                )
        except requests.exceptions.HTTPError as error:
            http_response: Union[Response, None] = error.response
            if isinstance(http_response, Response) and http_response.status_code in (404, 400):
                logger.error("Repository %s does not exist", repo_name)
                raise RepositoryNotFoundError(f" Repository {repo_name} does not exist")
            raise ArtifactoryError from error

    @overload
    def create_repo(self, repo: LocalRepository) -> LocalRepositoryResponse:
        ...

    @overload
    def create_repo(self, repo: VirtualRepository) -> VirtualRepositoryResponse:
        ...

    @overload
    def create_repo(self, repo: RemoteRepository) -> RemoteRepositoryResponse:
        ...

    @overload
    def create_repo(self, repo: FederatedRepository) -> FederatedRepositoryResponse:
        ...

    def create_repo(
        self,
        repo: AnyRepository,
    ) -> AnyRepositoryResponse:
        """
        Creates a local, virtual, remote or federated repository
        :param repo: Either a local, virtual, remote or federated repository
        :return: LocalRepositoryResponse, VirtualRepositoryResponse, RemoteRepositoryResponse
                 or FederatedRepositoryResponse object
        """
        repo_name = repo.key
        try:
            self.get_repo(repo_name)
            logger.error("Repository %s already exists", repo_name)
            raise RepositoryAlreadyExistsError(f"Repository {repo_name} already exists")
        except RepositoryNotFoundError:
            data = json.dumps(repo.model_dump(), default=custom_encoder)
            self._put(
                f"api/{self._uri}/{repo_name}",
                headers={"Content-Type": "application/json"},
                data=data,
            )
            logger.debug("Repository %s successfully created", repo_name)
            return self.get_repo(repo_name)

    @overload
    def update_repo(self, repo: LocalRepository) -> LocalRepositoryResponse:
        ...

    @overload
    def update_repo(self, repo: VirtualRepository) -> VirtualRepositoryResponse:
        ...

    @overload
    def update_repo(self, repo: RemoteRepository) -> RemoteRepositoryResponse:
        ...

    @overload
    def update_repo(self, repo: FederatedRepository) -> FederatedRepositoryResponse:
        ...

    def update_repo(
        self,
        repo: AnyRepository,
    ) -> AnyRepositoryResponse:
        """
        Updates a local, virtual or remote repository
        :param repo: Either a local, virtual or remote repository
        :return: LocalRepositoryResponse, VirtualRepositoryResponse or RemoteRepositoryResponse object
        """
        repo_name = repo.key
        self.get_repo(repo_name)

        repo_dict = json.dumps(repo.model_dump(exclude_unset=True), default=custom_encoder)

        self._post(
            f"api/{self._uri}/{repo_name}",
            headers={"Content-Type": "application/json"},
            data=repo_dict,
        )
        logger.debug("Repository %s successfully updated", repo_name)
        return self.get_repo(repo_name)

    # Remote repositories operations
    def list(self) -> List[SimpleRepository]:
        """
        Lists all the repositories
        :return: A list of repositories
        """
        response = self._get(f"api/{self._uri}")
        logger.debug("List all repositories successful")
        return [SimpleRepository(**repository) for repository in response.json()]

    def delete(self, repo_name: str) -> None:
        """
        Removes a local repository
        :param repo_name: Name of the repository to delete
        :return: None
        """

        self._delete(f"api/{self._uri}/{repo_name}")
        logger.debug("Repository %s successfully deleted", repo_name)



"""
Import all models here.
"""
from __future__ import annotations

from typing import Union

from .artifact import (
    ArtifactFileInfoResponse,
    ArtifactFolderInfoResponse,
    ArtifactInfoResponse,
    ArtifactPropertiesResponse,
    ArtifactStatsResponse,
)
from .auth import AccessTokenModel, ApiKeyModel, AuthModel, PasswordModel
from .build import (
    BuildAgent,
    BuildArtifact,
    BuildCreateRequest,
    BuildDeleteRequest,
    BuildDiffResponse,
    BuildDiffResponseDetail,
    BuildError,
    BuildInfo,
    BuildInfoDetail,
    BuildListResponse,
    BuildModules,
    BuildPromotionRequest,
    BuildPromotionResult,
    BuildProperties,
    BuildRun,
    Run,
    SimpleBuild,
)
from .group import Group, SimpleGroup
from .permission import Permission, PermissionV2, SimplePermission
from .repository import (
    FederatedRepository,
    FederatedRepositoryResponse,
    LocalRepository,
    LocalRepositoryResponse,
    RemoteRepository,
    RemoteRepositoryResponse,
    SimpleRepository,
    VirtualRepository,
    VirtualRepositoryResponse,
)
from .user import BaseUserModel, NewUser, SimpleUser, User, UserResponse

AnyRepositoryResponse = Union[
    LocalRepositoryResponse,
    VirtualRepositoryResponse,
    RemoteRepositoryResponse,
    FederatedRepositoryResponse,
]

AnyRepository = Union[LocalRepository, VirtualRepository, RemoteRepository, FederatedRepository]
AnyPermission = Union[Permission, PermissionV2]

__all__ = [
    "ArtifactFileInfoResponse",
    "ArtifactFolderInfoResponse",
    "ArtifactInfoResponse",
    "ArtifactPropertiesResponse",
    "ArtifactStatsResponse",
    "AccessTokenModel",
    "ApiKeyModel",
    "AuthModel",
    "PasswordModel",
    "Group",
    "SimpleGroup",
    "Permission",
    "PermissionV2",
    "SimplePermission",
    "LocalRepository",
    "LocalRepositoryResponse",
    "RemoteRepository",
    "RemoteRepositoryResponse",
    "FederatedRepository",
    "FederatedRepositoryResponse",
    "SimpleRepository",
    "VirtualRepository",
    "VirtualRepositoryResponse",
    "User",
    "UserResponse",
    "BaseUserModel",
    "NewUser",
    "SimpleUser",
    "AnyRepositoryResponse",
    "AnyRepository",
    "AnyPermission",
    "SimpleBuild",
    "BuildListResponse",
    "Run",
    "BuildRun",
    "BuildArtifact",
    "BuildModules",
    "BuildAgent",
    "BuildInfoDetail",
    "BuildInfo",
    "BuildPromotionResult",
    "BuildPromotionRequest",
    "BuildProperties",
    "BuildError",
    "BuildDeleteRequest",
    "BuildDiffResponseDetail",
    "BuildDiffResponse",
    "BuildCreateRequest",
]



"""
Definition of all utils.
"""
from __future__ import annotations

from typing import Any

from pydantic import SecretStr
from pydantic_core import to_jsonable_python


def custom_encoder(obj: Any) -> Any:
    """
    Custom encoder function to be passed to the default argument of json.dumps()
    :param obj: A pydantic object
    :return: An encoded pydantic object
    """
    if isinstance(obj, SecretStr):
        return obj.get_secret_value()
    return to_jsonable_python(obj)



from __future__ import annotations

import logging
from typing import List, Union

import requests
from requests import Response

from pyartifactory.exception import ArtifactoryError, UserAlreadyExistsError, UserNotFoundError
from pyartifactory.models.user import NewUser, SimpleUser, User, UserResponse
from pyartifactory.objects.object import ArtifactoryObject

logger = logging.getLogger("pyartifactory")


class ArtifactoryUser(ArtifactoryObject):
    """Models an artifactory user."""

    _uri = "security/users"

    def create(self, user: NewUser) -> UserResponse:
        """
        Create user
        :param user: NewUser object
        :return: User
        """
        username = user.name
        try:
            self.get(username)
            logger.error("User %s already exists", username)
            raise UserAlreadyExistsError(f"User {username} already exists")
        except UserNotFoundError:
            data = user.model_dump()
            data["password"] = user.password.get_secret_value()
            self._put(f"api/{self._uri}/{username}", json=data)
            logger.debug("User %s successfully created", username)
            return self.get(user.name)

    def get(self, name: str) -> UserResponse:
        """
        Read user from artifactory. Fill object if exist
        :param name: Name of the user to retrieve
        :return: UserModel
        """
        try:
            response = self._get(f"api/{self._uri}/{name}")
            logger.debug("User %s found", name)
            return UserResponse(**response.json())
        except requests.exceptions.HTTPError as error:
            http_response: Union[Response, None] = error.response
            if isinstance(http_response, Response) and http_response.status_code in (404, 400):
                logger.error("User %s does not exist", name)
                raise UserNotFoundError(f"{name} does not exist")
            raise ArtifactoryError from error

    def list(self) -> List[SimpleUser]:
        """
        Lists all the users
        :return: UserList
        """
        response = self._get(f"api/{self._uri}")
        logger.debug("List all users successful")
        return [SimpleUser(**user) for user in response.json()]

    def update(self, user: User) -> UserResponse:
        """
        Updates an artifactory user
        :param user: NewUser object
        :return: UserModel
        """
        username = user.name
        self.get(username)
        self._post(
            f"api/{self._uri}/{username}",
            json=user.model_dump(exclude={"lastLoggedIn", "realm"}),
        )
        logger.debug("User %s successfully updated", username)
        return self.get(username)

    def delete(self, name: str) -> None:
        """
        Remove user
        :param name: Name of the user to delete
        :return: None
        """
        self.get(name)
        self._delete(f"api/{self._uri}/{name}")
        logger.debug("User %s successfully deleted", name)

    def unlock(self, name: str) -> None:
        """
        Unlock user
        Even if the user doesn't exist, it succeed too
        :param name: Name of the user to unlock
        :return none
        """
        self._post(f"api/security/unlockUsers/{name}")
        logger.debug("User % successfully unlocked", name)



from __future__ import annotations

from typing import Optional, Tuple, Union

from pydantic import SecretStr

from pyartifactory.models.auth import AuthModel
from pyartifactory.objects.artifact import ArtifactoryArtifact
from pyartifactory.objects.build import ArtifactoryBuild
from pyartifactory.objects.group import ArtifactoryGroup
from pyartifactory.objects.permission import ArtifactoryPermission
from pyartifactory.objects.repository import ArtifactoryRepository
from pyartifactory.objects.security import ArtifactorySecurity
from pyartifactory.objects.user import ArtifactoryUser


class Artifactory:
    """Models artifactory."""

    def __init__(
        self,
        url: str,
        auth: Tuple[str, SecretStr],
        verify: Union[bool, str] = True,
        cert: Optional[str] = None,
        api_version: int = 1,
        timeout: Optional[int] = None,
    ):
        self.artifactory = AuthModel(
            url=url,
            auth=auth,
            verify=verify,
            cert=cert,
            api_version=api_version,
            timeout=timeout,
        )
        self.users = ArtifactoryUser(self.artifactory)
        self.groups = ArtifactoryGroup(self.artifactory)
        self.security = ArtifactorySecurity(self.artifactory)
        self.repositories = ArtifactoryRepository(self.artifactory)
        self.artifacts = ArtifactoryArtifact(self.artifactory)
        self.permissions = ArtifactoryPermission(self.artifactory)
        self.builds = ArtifactoryBuild(self.artifactory)



from __future__ import annotations

import logging
import os
import urllib
from collections.abc import Iterator
from pathlib import Path
from typing import Dict, List, Optional, Union

import requests
from pydantic import ValidationError
from requests import Response

from pyartifactory.exception import ArtifactNotFoundError, ArtifactoryError, BadPropertiesError, PropertyNotFoundError
from pyartifactory.models.artifact import (
    ArtifactFileInfoResponse,
    ArtifactFolderInfoResponse,
    ArtifactInfoResponse,
    ArtifactListResponse,
    ArtifactPropertiesResponse,
    ArtifactStatsResponse,
    Checksums,
)
from pyartifactory.objects.object import ArtifactoryObject

logger = logging.getLogger("pyartifactory")


class ArtifactoryArtifact(ArtifactoryObject):
    """Models an artifactory artifact."""

    def _walk(self, artifact_path: str, topdown: bool = True) -> Iterator[ArtifactInfoResponse]:
        """Iterate over artifact (file or directory) recursively.

        :param artifact_path: Path to file or folder in Artifactory
        :param topdown: True for a top-down (directory first) traversal
        """
        info = self.info(artifact_path)
        if not isinstance(info, ArtifactFolderInfoResponse):
            yield info
        else:
            if topdown:
                yield info
            for subdir in (child for child in info.children if child.folder is True):
                yield from self._walk(artifact_path + subdir.uri, topdown=topdown)
            for file in (child for child in info.children if child.folder is not True):
                yield from self._walk(artifact_path + file.uri, topdown=topdown)
            if not topdown:
                yield info

    def info(self, artifact_path: Union[Path, str]) -> ArtifactInfoResponse:
        """
        Retrieve information about a file or a folder

        See https://www.jfrog.com/confluence/display/JFROG/Artifactory+REST+API#ArtifactoryRESTAPI-FolderInfo
        and https://www.jfrog.com/confluence/display/JFROG/Artifactory+REST+API#ArtifactoryRESTAPI-FileInfo

        :param artifact_path: Path to file or folder in Artifactory
        """
        if isinstance(artifact_path, str):
            artifact_path = Path(artifact_path.lstrip("/"))

        try:
            artifact_as_posix = artifact_path.as_posix()
            artifact_as_url = urllib.parse.quote(artifact_as_posix)
            response = self._get(f"api/storage/{artifact_as_url}")
            try:
                artifact_info: ArtifactInfoResponse = ArtifactFolderInfoResponse.model_validate(response.json())
            except ValidationError:
                artifact_info = ArtifactFileInfoResponse.model_validate(response.json())
            return artifact_info
        except requests.exceptions.HTTPError as error:
            http_response: Union[Response, None] = error.response
            if isinstance(http_response, Response) and http_response.status_code == 404:
                logger.error("Artifact %s does not exist", artifact_path)
                raise ArtifactNotFoundError(f"Artifact {artifact_path} does not exist")
            raise ArtifactoryError from error

    def deploy(
        self,
        local_file_location: Union[Path, str],
        artifact_path: Union[Path, str],
        checksum_enabled: bool = False,
    ) -> ArtifactInfoResponse:
        """
        Deploy a file or directory.
        :param artifact_path: Path to artifactory in Artifactory
        :param local_file_location: Location of the file or folder to deploy
        :param checksum_enabled: Enable checksum generation and use it for validation of the deployment
        """
        local_file = Path(local_file_location)
        artifact_folder = Path(artifact_path)

        if local_file.is_dir():
            for root, _, files in os.walk(local_file.as_posix()):
                new_root = f"{artifact_folder}/{root[len(local_file.as_posix()):]}"
                for file in files:
                    self.deploy(Path(f"{root}/{file}"), Path(f"{new_root}/{file}"), checksum_enabled)
        else:
            if checksum_enabled:
                artifact_check_sums = Checksums.generate(local_file)
                try:
                    self._put(
                        route=artifact_folder.as_posix(),
                        headers={
                            "X-Checksum-Deploy": "true",
                            "X-Checksum-Sha1": artifact_check_sums.sha1,
                            "X-Checksum-Sha256": artifact_check_sums.sha256,
                            "X-Checksum": artifact_check_sums.md5,
                        },
                    )
                except requests.exceptions.HTTPError as error:
                    if error.response.status_code == 404:
                        message = (
                            f"Artifact {artifact_path} does not exist in Artifactory,"
                            f" content is expected to deploy by checksum"
                        )
                        logger.error(message)
                        raise ArtifactNotFoundError(message)
                    raise ArtifactoryError from error
            else:
                with local_file.open("rb") as stream:
                    self._put(route=artifact_folder.as_posix(), data=stream)

            logger.debug("Artifact %s successfully deployed", local_file)
        return self.info(artifact_folder)

    @staticmethod
    def _get_path_prefix(artifact_path: str):
        if artifact_path.startswith("/"):
            artifact_path = artifact_path[1:]
        return artifact_path.rsplit("/", 1)[0] + "/" if "/" in artifact_path else ""

    @staticmethod
    def _remove_prefix(_str: str, prefix: str) -> str:
        if _str.startswith(prefix):
            return _str[len(prefix) :]
        raise ValueError(f"Input string, '{_str}', doesn't have the prefix: '{prefix}'")

    def _download(self, artifact_path: str, local_directory_path: Optional[Path] = None) -> Path:
        """
        Download artifact (file) into local directory.
        :param artifact_path: Path to file in Artifactory
        :param local_directory_path: Local path to where the artifact will be downloaded
        :return: File name
        """
        artifact_path = artifact_path.lstrip("/")
        local_filename = artifact_path.split("/")[-1]

        if local_directory_path:
            local_directory_path.mkdir(parents=True, exist_ok=True)
            local_file_full_path = local_directory_path / local_filename
        else:
            local_file_full_path = Path(local_filename)

        artifact_path_url = urllib.parse.quote(artifact_path)
        with self._get(f"{artifact_path_url}", stream=True) as response, local_file_full_path.open("wb") as file:
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:  # filter out keep-alive new chunks
                    file.write(chunk)
        logger.debug("Artifact %s successfully downloaded", local_filename)
        return local_file_full_path

    def download(self, artifact_path: str, local_directory_path: str = ".") -> Path:
        """
        Download artifact (file or directory) into local directory.
        :param artifact_path: Path to file or directory in Artifactory
        :param local_directory_path: Local path to where the artifact will be downloaded
        :return: File name
        """
        artifact_path = artifact_path.rstrip("/")
        basename = artifact_path.split("/")[-1]
        prefix = self._get_path_prefix(artifact_path)
        for art in self._walk(artifact_path):
            full_path = art.repo + art.path
            local_artifact_path = self._remove_prefix(full_path, prefix)
            local_path = Path(local_directory_path) / local_artifact_path
            if isinstance(art, ArtifactFolderInfoResponse):
                local_path.mkdir(exist_ok=True)
            else:
                self._download(full_path, local_path.parent)
        return Path(local_directory_path).joinpath(basename)

    def list(
        self,
        artifact_path: str,
        recursive: bool = True,
        depth: Optional[int] = None,
        list_folders: bool = True,
    ) -> ArtifactListResponse:
        """
        Retrieve a list of files or a folders

        See https://www.jfrog.com/confluence/display/JFROG/Artifactory+REST+API#ArtifactoryRESTAPI-FileList

        :param artifact_path: Path to folder in Artifactory
        :param recursive: Recursively retrieve files and folders
        :param depth: The depth to recursively retrieve
        :param list_folders: Whether or not to include folders in the response
        :return: A list of artifacts in Artifactory
        """
        try:
            params = {
                "deep": int(recursive),
                "listFolders": int(list_folders),
            }
            if depth is not None:
                params.update(depth=depth)
            response = self._get(f"api/storage/{artifact_path}?list", params=params)
            artifact_list: ArtifactListResponse = ArtifactListResponse.model_validate(response.json())
            return artifact_list
        except requests.exceptions.HTTPError as error:
            http_response: Union[Response, None] = error.response
            if isinstance(http_response, Response) and http_response.status_code == 404:
                logger.error("Artifact %s does not exist", artifact_path)
                raise ArtifactNotFoundError(f"Artifact {artifact_path} does not exist")
            raise ArtifactoryError from error

    def properties(self, artifact_path: str, properties: Optional[List[str]] = None) -> ArtifactPropertiesResponse:
        """
        :param artifact_path: Path to file in Artifactory
        :param properties: List of properties to retrieve
        :return: Artifact properties
        """
        if properties is None:
            properties = []
        artifact_path = artifact_path.lstrip("/")
        try:
            response = self._get(
                f"api/storage/{artifact_path}",
                params={"properties": ",".join(properties)},
            )
            logger.debug("Artifact Properties successfully retrieved")
            return ArtifactPropertiesResponse(**response.json())
        except requests.exceptions.HTTPError as error:
            http_response: Union[Response, None] = error.response
            if isinstance(http_response, Response) and http_response.status_code == 404:
                raise PropertyNotFoundError(f"Properties {properties} were not found on artifact {artifact_path}")
            raise ArtifactoryError from error

    def set_properties(
        self,
        artifact_path: str,
        properties: Dict[str, List[str]],
        recursive: bool = True,
    ) -> ArtifactPropertiesResponse:
        """
        :param artifact_path: Path to file or folder in Artifactory
        :param properties: List of properties to update
        :param recursive: If set to true, properties will be applied recursively to subfolders and files
        :return: None
        """
        if properties is None:
            properties = {}
        artifact_path = artifact_path.lstrip("/")
        properties_param_str = ""
        for k, v in properties.items():
            values_str = ",".join(v)
            properties_param_str += f"{k}={values_str};"
        try:
            self._put(
                f"api/storage/{artifact_path}",
                params={
                    "recursive": int(recursive),
                    "properties": properties_param_str,
                },
            )
            logger.debug("Artifact Properties successfully set")
            return self.properties(artifact_path)
        except requests.exceptions.HTTPError as error:
            http_response: Union[Response, None] = error.response
            if isinstance(http_response, Response) and http_response.status_code == 404:
                logger.error("Artifact %s does not exist", artifact_path)
                raise ArtifactNotFoundError(f"Artifact {artifact_path} does not exist")
            if isinstance(http_response, Response) and http_response.status_code == 400:
                logger.error("A property value includes forbidden special characters")
                raise BadPropertiesError("A property value includes forbidden special characters")
            raise ArtifactoryError from error

    def update_properties(
        self,
        artifact_path: str,
        properties: Dict[str, List[str]],
        recursive: bool = True,
    ) -> ArtifactPropertiesResponse:
        """
        :param artifact_path: Path to file or folder in Artifactory
        :param properties: List of properties to update
        :param recursive: If set to true, properties will be applied recursively to subfolders and files
        :return: None
        """
        if properties is None:
            properties = {}
        artifact_path = artifact_path.lstrip("/")
        try:
            self._patch(
                f"api/metadata/{artifact_path}",
                params={"recursiveProperties": int(recursive)},
                headers={"Content-Type": "application/json"},
                json={"props": properties},
            )
            logger.debug("Artifact Properties successfully updated")
            return self.properties(artifact_path)
        except requests.exceptions.HTTPError as error:
            http_response: Union[Response, None] = error.response
            if isinstance(http_response, Response) and http_response.status_code == 400:
                logger.error("Error updating artifact properties")
                raise ArtifactoryError("Error updating artifact properties")
            raise ArtifactoryError from error

    def stats(self, artifact_path: str) -> ArtifactStatsResponse:
        """
        :param artifact_path: Path to file in Artifactory
        :return: Artifact Stats
        """
        artifact_path = artifact_path.lstrip("/")
        response = self._get(f"api/storage/{artifact_path}?stats")
        logger.debug("Artifact stats successfully retrieved")
        return ArtifactStatsResponse(**response.json())

    def copy(self, artifact_current_path: str, artifact_new_path: str, dryrun: bool = False) -> ArtifactInfoResponse:
        """
        :param artifact_current_path: Current path to file
        :param artifact_new_path: New path to file
        :param dryrun: Dry run
        :return: ArtifactInfoResponse: info of the copied artifact
        """
        artifact_current_path = artifact_current_path.lstrip("/")
        artifact_new_path = artifact_new_path.lstrip("/")
        dry = 1 if dryrun else 0

        self._post(f"api/copy/{artifact_current_path}?to={artifact_new_path}&dry={dry}")
        logger.debug("Artifact %s successfully copied", artifact_current_path)
        return self.info(artifact_new_path)

    def move(self, artifact_current_path: str, artifact_new_path: str, dryrun: bool = False) -> ArtifactInfoResponse:
        """
        :param artifact_current_path: Current path to file
        :param artifact_new_path: New path to file
        :param dryrun: Dry run
        :return: ArtifactInfoResponse: info of the moved artifact
        """
        artifact_current_path = artifact_current_path.lstrip("/")
        artifact_new_path = artifact_new_path.lstrip("/")

        dry = 1 if dryrun else 0

        self._post(f"api/move/{artifact_current_path}?to={artifact_new_path}&dry={dry}")
        logger.debug("Artifact %s successfully moved", artifact_current_path)
        return self.info(artifact_new_path)

    def delete(self, artifact_path: str) -> None:
        """
        :param artifact_path: Path to file in Artifactory
        :return: bool
        """
        artifact_path = artifact_path.lstrip("/")
        self._delete(f"{artifact_path}")
        logger.debug("Artifact %s successfully deleted", artifact_path)



from __future__ import annotations

import logging
from typing import List, Union

import requests
from requests import Response

from pyartifactory.exception import ArtifactoryError, GroupAlreadyExistsError, GroupNotFoundError
from pyartifactory.models.group import Group
from pyartifactory.objects.object import ArtifactoryObject

logger = logging.getLogger("pyartifactory")


class ArtifactoryGroup(ArtifactoryObject):
    """Models artifactory groups."""

    _uri = "security/groups"

    def create(self, group: Group) -> Group:
        """
        Creates a new group in Artifactory or replaces an existing group
        :param group: Group to create
        :return: Created group
        """
        group_name = group.name
        try:
            self.get(group_name)
            logger.error("Group %s already exists", group_name)
            raise GroupAlreadyExistsError(f"Group {group_name} already exists")
        except GroupNotFoundError:
            self._put(f"api/{self._uri}/{group_name}", json=group.model_dump())
            logger.debug("Group %s successfully created", group_name)
            return self.get(group.name)

    def get(self, name: str) -> Group:
        """
        Get the details of an Artifactory Group
        :param name: Name of the group to retrieve
        :return: Found artifactory group
        """
        try:
            response = self._get(f"api/{self._uri}/{name}", params={"includeUsers": True})
            logger.debug("Group %s found", name)
            return Group(**response.json())
        except requests.exceptions.HTTPError as error:
            http_response: Union[Response, None] = error.response
            if isinstance(http_response, Response) and http_response.status_code in (404, 400):
                logger.error("Group %s does not exist", name)
                raise GroupNotFoundError(f"Group {name} does not exist")
            raise ArtifactoryError from error

    def list(self) -> List[Group]:
        """
        Lists all the groups
        :return: GroupList
        """
        response = self._get(f"api/{self._uri}")
        logger.debug("List all groups successful")
        return [Group(**group) for group in response.json()]

    def update(self, group: Group) -> Group:
        """
        Updates an exiting group in Artifactory with the provided group details.
        :param group: Group to be updated
        :return: Updated group
        """
        group_name = group.name
        self.get(group_name)
        self._post(f"api/{self._uri}/{group_name}", json=group.model_dump())
        logger.debug("Group %s successfully updated", group_name)
        return self.get(group_name)

    def delete(self, name: str) -> None:
        """
        Removes a group
        :param name: Name of the group to delete
        :return: None
        """
        self.get(name)
        self._delete(f"api/{self._uri}/{name}")
        logger.debug("Group %s successfully deleted", name)



"""
Definition of all auth models.
"""
from __future__ import annotations

from typing import Optional, Tuple, Union

from pydantic import BaseModel, SecretStr


class AuthModel(BaseModel):
    """Models an auth response."""

    url: str
    auth: Tuple[str, SecretStr]
    verify: Union[bool, str] = True
    cert: Optional[str] = None
    api_version: int = 1
    timeout: Optional[int] = None


class ApiKeyModel(BaseModel):
    """Models an api key."""

    apiKey: SecretStr


class PasswordModel(BaseModel):
    """Models a password."""

    password: SecretStr


class AccessTokenModel(BaseModel):
    """Model an access token."""

    access_token: str
    expires_in: Optional[int] = 3600
    scope: str
    refresh_token: Optional[str] = None
    token_type: str



"""
Definition of all group models.
"""
from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel


class Group(BaseModel):
    """Models a group."""

    name: str
    description: Optional[str] = None
    autoJoin: bool = False
    adminPrivileges: bool = False
    realm: Optional[str] = None
    realmAttributes: Optional[str] = None
    watchManager: bool = False
    policyManager: bool = False
    userNames: Optional[List[str]] = None


class SimpleGroup(BaseModel):
    """Models a simple group."""

    name: str
    uri: str
